/**********************************************
                SYNTAX GUIDE

On the next lines "stack" means a LIFO stack; only integers allowed

  <ANY NUMBER>                Pushes number into stack
  POP                         Pops number from stack
  PRT                         Prints stack top
  LBL <NUMBER>                Creates a label, numbered <NUMBER>, pointing to the next instruction; does not affect stack, in spite of the number
  GTO <NUMBER>                Continue execution from the intruction defined by label <NUMBER>, if it exists; otherwise stops with error
  IF0GTO <NUMBER>             If stack top equals 0 (zero), jumbs to the instruction labeled <NUMBER>; otherwise continues execution
  IFN0GTO <NUMBER>            If stack top differs from 0 (zero), jumbs to the instruction labeled <NUMBER>; otherwise continues execution
  LET <NON NUMBER>            Stores the stack top into the varibale named <NON NUMBER>
  ALET <NON NUMBER> <NUMBER>
  INP                         Reads input data and pushes it into stack; from command line, input data is in the end of program
  + - . / %                   Performs the aritmetical operation (<OP>) between the top of the stack (<TOP>) and the preceeding number (<PREC>),
                              in the order <PREC> <OP> <TOP>. The '.' stands for multiplication
  <NON NUMBER>                Pushes the content of the variable <NON NUMBER> into stack, if it exists; if it does not exist and if <NON NUMBER>
                              is an array name, replaces the top of the stack top with the value of the array item numbered by the current top of the
                              stack; otherwise stops with error

************************************************/

#include <ctype.h>      // isdigit
#include <string.h>    // strcpy
#include <stdio.h>    // printf
#include <stdlib.h>    // malloc, free
#include <unistd.h>   // fork, exec
#include <sys/wait.h> // wait
#include "1_main.h"
#include "do_the_fork.cpp"


int main(int argc, char* argv[]) {

  // carrega a linguagem
  TOKEN_t *lang_array = NULL;
  int lang_len = 0;
  init_language(&lang_array, lang_len);

  // mostra a linguagem
//  for (int i = 0; i < lang_len; i++) {
//    if (strcmp(lang_array[i].name, "") == 0) {
//      if (lang_array[i].has_named_label) {
//        printf("<VARNAME>\n");
//      } else {
//        printf("<NUMBER>\n");
//      }
//    } else {
//      printf("%s ", lang_array[i].name);
//      if (lang_array[i].elements > 1) {
//        if (lang_array[i].has_named_label) {
//          printf("<VARNAME> ");
//        } else {
//          printf("<NUMBER> ");
//        }
//        if (lang_array[i].elements > 2) {
//          printf("<NUMBER>");
//        }
//      }
//      printf("\n");
//    }
//  }

  // procura o programa
  RANDOM_t *r_num_passos = new RANDOM_t(1);
  RANDOM_t *r_variable_name = new RANDOM_t(3);
  RANDOM_t *r_a_number = new RANDOM_t(4);

  RANDOM_t *r_passo = new RANDOM_t(2);
  r_passo->criaDist(0, lang_len - 1);

  RANDOM_t *r_toss_coin = new RANDOM_t(5);
  r_toss_coin->criaDist(0, 1);

  RANDOM_t *r_rand = new RANDOM_t(8);

  // CRITERIOS ESTETICOS
  r_num_passos->criaDist(4, 8);
  r_variable_name->criaDist(0, 4);
  r_a_number->criaDist(0, 9);
  //

  int *tokenlist = NULL;
  int tokenlen = 0;

  int *varlist = NULL;
  int varlen = 0;

  int *lbllist = NULL;
  int lbllen = 0;

  char **programa = NULL;
  int programa_len = 0;

  int num = 1000;
  while (num) {

    // sorteia uma programa
    int num_passos = r_num_passos->aleatorio();
    if (num_passos > tokenlen) {
      if (tokenlist) free(tokenlist);
      tokenlist = (int *)malloc(num_passos * sizeof(int));
      tokenlen = num_passos;
    }

    if (programa) {
      for (int i = 0; i < programa_len; i++) free(programa[i]);
      free(programa);
    }
    programa = NULL;
    programa_len = 0;

    // sorteia cada passo
    for (int i = 0; i < num_passos; i++) {
      int token = r_passo->aleatorio();

      // checa se é um número ou uma variável
      if (strcmp(lang_array[token].name, "") == 0) {
        if (lang_array[token].has_named_label) {
          // é uma variável que se coloca neste lugar: pode existir ou não
          bool ok = false;
          if (r_toss_coin->aleatorio()) {
            // pega uma existente
            if (varlen != 0) {
              ok = true;
              if (varlen > 1) {
                r_rand->criaDist(0, varlen - 1);
                programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
                programa[programa_len] = (char *)malloc(sizeof(char) * 2);
                (*(programa + programa_len))[0] = 'a' + varlist[r_rand->aleatorio()];
                (*(programa + programa_len))[1] = '\0';
                programa_len++;
              } else {
                programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
                programa[programa_len] = (char *)malloc(sizeof(char) * 2);
                (*(programa + programa_len))[0] = 'a' + varlist[0];
                (*(programa + programa_len))[1] = '\0';
                programa_len++;
              }
            }
          }
          if (! ok) {
            // tem de tentar criar a variável através de sorteio. Só acrescenta à lista se sortear uma nova
            int srt = r_variable_name->aleatorio();
            bool achou = false;
            for (int i = 0; i < varlen; i++) {
              if (varlist[i] == srt) {
                achou = true;
                break;
              }
            }
            if (! achou) {
              varlist = (int *)realloc(varlist, (varlen + 1) * sizeof(int));
              varlist[varlen] = srt;
              varlen++;
            }
            programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
            programa[programa_len] = (char *)malloc(sizeof(char) * 2);
            (*(programa + programa_len))[0] = 'a' + srt;
            (*(programa + programa_len))[1] = '\0';
            programa_len++;
          }
        } else {
          // é um número que se coloca neste lugar
          int srt = r_a_number->aleatorio();
          programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
          programa[programa_len] = NULL;

          int aux = srt;
          char caux;
          int wordlen = 0;
          while (aux) {
            caux = '0' + (aux % 10);
            programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
            (*(programa + programa_len))[wordlen] = caux;
            wordlen++;
            aux /= 10;
          }
          programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
          (*(programa + programa_len))[wordlen] = '\0';

          programa_len++;
        }
      } else {
        // trata-se de um comando
        programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
        programa[programa_len] = (char *)malloc(sizeof(char) * (1 + strlen(lang_array[token].name)));
        programa[programa_len] = strcpy(programa[programa_len], lang_array[token].name);
        programa_len++;

        //  vejamos os elementos
        if (lang_array[token].elements > 1) {
          if (lang_array[token].has_named_label) {
            // coloca o nome de uma variável

            bool ok = false;
            if (r_toss_coin->aleatorio()) {
              // pega uma existente
              if (varlen != 0) {
                ok = true;
                if (varlen > 1) {
                  r_rand->criaDist(0, varlen - 1);
                  programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
                  programa[programa_len] = (char *)malloc(sizeof(char) * 2);
                  (*(programa + programa_len))[0] = 'a' + varlist[r_rand->aleatorio()];
                  (*(programa + programa_len))[1] = '\0';
                  programa_len++;
                } else {
                  programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
                  programa[programa_len] = (char *)malloc(sizeof(char) * 2);
                  (*(programa + programa_len))[0] = 'a' + varlist[0];
                  (*(programa + programa_len))[1] = '\0';
                  programa_len++;
                }
              }
            }
            if (! ok) {
              // tem de tentar criar a variável através de sorteio. Só acrescenta à lista se sortear uma nova
              int srt = r_variable_name->aleatorio();
              bool achou = false;
              for (int i = 0; i < varlen; i++) {
                if (varlist[i] == srt) {
                  achou = true;
                  break;
                }
              }
              if (! achou) {
                varlist = (int *)realloc(varlist, (varlen + 1) * sizeof(int));
                varlist[varlen] = srt;
                varlen++;
              }
              programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
              programa[programa_len] = (char *)malloc(sizeof(char) * 2);
              (*(programa + programa_len))[0] = 'a' + srt;
              (*(programa + programa_len))[1] = '\0';
              programa_len++;
            }

          } else {
            // coloca um número que é um label

            bool ok = false;
            if (r_toss_coin->aleatorio()) {

              // pega um existente
              if (lbllen != 0) {
                ok = true;
                if (lbllen > 1) {
                  r_rand->criaDist(0, lbllen - 1);
                  programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
                  programa[programa_len] = NULL;

                  int aux = lbllist[r_rand->aleatorio()];
                  char caux;
                  int wordlen = 0;
                  while (aux) {
                    caux = '0' + (aux % 10);
                    programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
                    (*(programa + programa_len))[wordlen] = caux;
                    wordlen++;
                    aux /= 10;
                  }
                  programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
                  (*(programa + programa_len))[wordlen] = '\0';

                  programa_len++;
                } else {
                  programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
                  programa[programa_len] = NULL;

                  int aux = lbllist[0];
                  char caux;
                  int wordlen = 0;
                  while (aux) {
                    caux = '0' + (aux % 10);
                    programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
                    (*(programa + programa_len))[wordlen] = caux;
                    wordlen++;
                    aux /= 10;
                  }
                  programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
                  (*(programa + programa_len))[wordlen] = '\0';

                  programa_len++;
                }
              }
            }
            if (! ok) {
              // tem de tentar criar o label através de sorteio. Só acrescenta à lista se sortear um novo
              int srt = r_a_number->aleatorio();
              bool achou = false;
              for (int i = 0; i < lbllen; i++) {
                if (lbllist[i] == srt) {
                  achou = true;
                  break;
                }
              }
              if (! achou) {
                lbllist = (int *)realloc(lbllist, (lbllen + 1) * sizeof(int));
                lbllist[lbllen] = srt;
                lbllen++;
              }
              programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
              programa[programa_len] = NULL;

              int aux = srt;
              char caux;
              int wordlen = 0;
              while (aux) {
                caux = '0' + (aux % 10);
                programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
                (*(programa + programa_len))[wordlen] = caux;
                wordlen++;
                aux /= 10;
              }
              programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
              (*(programa + programa_len))[wordlen] = '\0';

              programa_len++;
            }
          }
          if (lang_array[token].elements > 2) {
            // coloca um número que é um índice de um array

            int srt = r_a_number->aleatorio();
            programa = (char **)realloc(programa, (programa_len + 1) * sizeof(char *));
            programa[programa_len] = NULL;

            int aux = srt;
            char caux;
            int wordlen = 0;
            while (aux) {
              caux = '0' + (aux % 10);
              programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
              (*(programa + programa_len))[wordlen] = caux;
              wordlen++;
              aux /= 10;
            }
            programa[programa_len] = (char *)realloc(programa[programa_len], sizeof(char) * (1 + wordlen));
            (*(programa + programa_len))[wordlen] = '\0';

            programa_len++;
          }
        }
      }
    }

    // acrescenta os valores finais, a somar
    programa = (char **)realloc(programa, (programa_len + 2) * sizeof(char *));
    programa[programa_len] = (char *)malloc(sizeof(char) * 2);
    programa[programa_len] = strcpy(programa[programa_len], "5");
    programa_len++;
    programa[programa_len] = (char *)malloc(sizeof(char) * 2);
    programa[programa_len] = strcpy(programa[programa_len], "2");
    programa_len++;


    // chama o programa!
    int status;
    char *result = NULL;
    result = (char *)malloc(GAZ_BUFFSIZE * sizeof(char));
    bool ret = false;
    ret = do_the_fork(programa_len, programa, &status, &result);

    if (!ret) {
      //printf("Deu pau\n");
    } else {
      if (status != 0) {
        //printf("Deu mal, %d\n", WEXITSTATUS(status));
      } else {
        //printf("Deu bem, '%s': %d\n", result, atoi(result));
        if (atoi(result) == 7) {
          printf("Deu bem, '%s': %d, num = %d\n", result, atoi(result), num);
          // mostra o programa
          for(int i = 0; i < programa_len; i++) {
            printf("%s ", programa[i]);
          }
          printf("\n");
          break;
        }
      }
    }
    free(result);

    num--;
  }

  // limpa memória
  for (int i = 0; i < lang_len; i++) free(lang_array[i].name);
  if (lang_array) free(lang_array);

  // deleta os rand
  delete r_num_passos;
  delete r_variable_name;
  delete r_a_number;

  delete r_passo;

  delete r_toss_coin;

  delete r_rand;


//  for (int i = 0; i < argc; i++) {
//    free(args[i]);
//    args[i] = NULL;
//  }
//  free(args[argc]);
//  free(args);

  return 0;
}
