/**********************************************
                SYNTAX GUIDE

On the next lines "stack" means a LIFO stack; only integers allowed

  <ANY NUMBER>                Pushes number into stack
  POP                         Pops number from stack
  PRT                         Prints stack top
  LBL <NUMBER>                Creates a label, numbered <NUMBER>, pointing to the next instruction; does not affect stack, in spite of the number
  GTO <NUMBER>                Continue execution from the intruction defined by label <NUMBER>, if it exists; otherwise stops with error
  IF0GTO <NUMBER>             If stack top equals 0 (zero), jumbs to the instruction labeled <NUMBER>; otherwise continues execution
  IFN0GTO <NUMBER>            If stack top differs from 0 (zero), jumbs to the instruction labeled <NUMBER>; otherwise continues execution
  LET <NON NUMBER>            Stores the stack top into the varibale named <NON NUMBER>
  ALET <NON NUMBER> <NUMBER>
  INP                         Reads input data and pushes it into stack; from command line, input data is in the end of program
  + - . / %                   Performs the aritmetical operation (<OP>) between the top of the stack (<TOP>) and the preceeding number (<PREC>),
                              in the order <PREC> <OP> <TOP>. The '.' stands for multiplication
  <NON NUMBER>                Pushes the content of the variable <NON NUMBER> into stack, if it exists; if it does not exist and if <NON NUMBER>
                              is an array name, replaces the top of the stack top with the value of the array item numbered by the current top of the
                              stack; otherwise stops with error

************************************************/

struct TOKEN_t {
  int elements;
  char *name;
  bool has_named_label;
};

void init_language(TOKEN_t **lang_array, int &lang_len) {
  int TOKEN_NUM = 16;

  *lang_array = (TOKEN_t *)malloc(TOKEN_NUM * sizeof(TOKEN_t));

  (*lang_array)[0].name = (char *)malloc(4 * sizeof(char));
  strcpy((*lang_array)[0].name, "POP");
  (*lang_array)[0].elements = 1;
  (*lang_array)[0].has_named_label = false;

  (*lang_array)[1].name = (char *)malloc(4 * sizeof(char));
  strcpy((*lang_array)[1].name, "PRT");
  (*lang_array)[1].elements = 1;
  (*lang_array)[1].has_named_label = false;

  (*lang_array)[2].name = (char *)malloc(4 * sizeof(char));
  strcpy((*lang_array)[2].name, "LBL");
  (*lang_array)[2].elements = 2;
  (*lang_array)[2].has_named_label = false;

  (*lang_array)[3].name = (char *)malloc(4 * sizeof(char));
  strcpy((*lang_array)[3].name, "GTO");
  (*lang_array)[3].elements = 2;
  (*lang_array)[3].has_named_label = false;

  (*lang_array)[4].name = (char *)malloc(7 * sizeof(char));
  strcpy((*lang_array)[4].name, "IF0GTO");
  (*lang_array)[4].elements = 2;
  (*lang_array)[4].has_named_label = false;

  (*lang_array)[5].name = (char *)malloc(8 * sizeof(char));
  strcpy((*lang_array)[5].name, "IFN0GTO");
  (*lang_array)[5].elements = 2;
  (*lang_array)[5].has_named_label = false;

  (*lang_array)[6].name = (char *)malloc(4 * sizeof(char));
  strcpy((*lang_array)[6].name, "LET");
  (*lang_array)[6].elements = 2;
  (*lang_array)[6].has_named_label = true;

  (*lang_array)[7].name = (char *)malloc(5 * sizeof(char));
  strcpy((*lang_array)[7].name, "ALET");
  (*lang_array)[7].elements = 3;
  (*lang_array)[7].has_named_label = true;

  (*lang_array)[8].name = (char *)malloc(4 * sizeof(char));
  strcpy((*lang_array)[8].name, "INP");
  (*lang_array)[8].elements = 1;
  (*lang_array)[8].has_named_label = false;

  (*lang_array)[9].name = (char *)malloc(2 * sizeof(char));
  strcpy((*lang_array)[9].name, "+");
  (*lang_array)[9].elements = 1;
  (*lang_array)[9].has_named_label = false;

  (*lang_array)[10].name = (char *)malloc(2 * sizeof(char));
  strcpy((*lang_array)[10].name, "-");
  (*lang_array)[10].elements = 1;
  (*lang_array)[10].has_named_label = false;

  (*lang_array)[11].name = (char *)malloc(2 * sizeof(char));
  strcpy((*lang_array)[11].name, ".");
  (*lang_array)[11].elements = 1;
  (*lang_array)[11].has_named_label = false;

  (*lang_array)[12].name = (char *)malloc(2 * sizeof(char));
  strcpy((*lang_array)[12].name, "/");
  (*lang_array)[12].elements = 1;
  (*lang_array)[12].has_named_label = false;

  (*lang_array)[13].name = (char *)malloc(2 * sizeof(char));
  strcpy((*lang_array)[13].name, "%");
  (*lang_array)[13].elements = 1;
  (*lang_array)[13].has_named_label = false;

  // variável
  (*lang_array)[14].name = (char *)malloc(1 * sizeof(char));
  strcpy((*lang_array)[14].name, "");
  (*lang_array)[14].elements = 1;
  (*lang_array)[14].has_named_label = true;

  // número
  (*lang_array)[15].name = (char *)malloc(1 * sizeof(char));
  strcpy((*lang_array)[15].name, "");
  (*lang_array)[15].elements = 1;
  (*lang_array)[15].has_named_label = false;

  lang_len = TOKEN_NUM;
  return;
}

bool do_the_fork(int argc, char* argv[], int *return_status = NULL, char **le_result = NULL) {
  bool retval = false;

  int pipe_fd[2];   // pipe descriptor
  char pipe_buffer; // pipe buffer

  pipe(pipe_fd);      // open pipe
  pid_t pid = fork();
  //printf("- right after fork(), pid = %d\n", pid);
  if (pid == 0) {
    //sleep(1);
    int exit_status = 0;

    // close read-end of pipe
    close(pipe_fd[0]);

//    printf("\tSubprg: recebeu %d argumentos:", argc);
//    for (int i = 0; i < argc; i++) {
//      if (i > 0) printf(",");
//      printf(" %s", argv[i]);
//    }
//    printf("\n");

    int MAX_IT = 10000; // CRITERIO ESTETICO
    int iteration_counter = 0; // iteration counter

    int rescount = 1 ;       // results count
    int restop = 0;          // results top
    int *restack = NULL;     // results stack
    restack = (int *)malloc(sizeof(int));

    int varcount = 0;        // variables count
    char** varnstack = NULL; // variables names stack
    int *varstack = NULL;    // variables values stack

    int arrcount = 0;        // arrays count
    int *arrcounter = NULL;  // array counter (number of elementes, max, in each array)
    char **arrnstack = NULL; // array labels stack
    int **arrstack = NULL;   // array values stack

    int lblcount = 0;        // labels count
    int *lblnstack = NULL;   // labels names stack
    int *lblstack = NULL;    // labels pointers stack

    int inputpointer = argc - 1; // input pointer

    // carrega todos os labels
    for (int i = 0; i < argc; i++) {
      if (strcmp(argv[i], "LBL") == 0) {
        lblnstack = (int *)realloc(lblnstack, (lblcount + 1) * sizeof(int));
        lblnstack[lblcount] = atoi(argv[i + 1]);

        lblstack = (int *)realloc(lblstack, (lblcount + 1) * sizeof(int));
        lblstack[lblcount] = i + 2;
        lblcount++;
      }
    }

    int pointer = 1;         // program pointer
    while (pointer < argc) {

      iteration_counter++;
      if (iteration_counter > MAX_IT) {
        //printf("Max iteration reached\n");
        exit_status = -3;
        break;
      }

      // números são guardados no stack de resultados
      if ( is_integer(argv[pointer]) ) {
        if (restop >= rescount - 1) {
          rescount = restop + 2;
          restack = (int *)realloc(restack, rescount * sizeof(int));
        }
        restop++;
        restack[restop] = atoi(argv[pointer]);

      // comando: input
      } else if (strcmp(argv[pointer], "INP") == 0) {
        if (restop >= rescount - 1) {
          rescount = restop + 2;
          restack = (int *)realloc(restack, rescount * sizeof(int));
        }
        restop++;
        restack[restop] = atoi(argv[inputpointer]);
        inputpointer--;

      // comando: define label
      } else if (strcmp(argv[pointer], "LBL") == 0) {
        // pula o label propriamente dito
        //restop--;
        pointer++;
      // comando: gto
      } else if (strcmp(argv[pointer], "GTO") == 0) {

        // procura o label na pilha de nomes de label
        bool achou = false;
        for (int i = 0; i < lblcount; i++) {
          if (atoi(argv[pointer + 1]) == lblnstack[i]) {
            pointer = lblstack[i];
            achou = true;
            break;
          }
        }
        if (achou) {
          //restop--;
          continue;
        }

        // só chega aqui se não achou o rótulo. Erro, né?
        //printf("Label not found: %s\n", argv[pointer + 1]);
        exit_status = -1;
        break;

      // comando: ifn0gto
      } else if (strcmp(argv[pointer], "IFN0GTO") == 0) {
        // checa se é zero
        if (restack[restop] == 0) {
          // pula o próximo, que deve ser o label do goto
          pointer++;
        } else {
          // vai para o label

          // procura o label na pilha de nomes de label
          int lbl = atoi(argv[pointer + 1]);
          bool achou = false;
          for (int i = 0; i < lblcount; i++) {
            if (lbl == lblnstack[i]) {
              pointer = lblstack[i];
              achou = true;
              break;
            }
          }
          if (achou) {
            continue;
          } else {
            //printf("Label not found: %d\n", lbl);
            exit(-1);
          }
        }

      // comando: if0gto
      } else if (strcmp(argv[pointer], "IF0GTO") == 0) {
        // checa se é != zero
        if (restack[restop] != 0) {
          // pula o próximo, que deve ser o label do goto
          pointer++;
        } else {
          // vai para o label

          // procura o label na pilha de nomes de label
          int lbl = atoi(argv[pointer + 1]);
          bool achou = false;
          for (int i = 0; i < lblcount; i++) {
            if (lbl == lblnstack[i]) {
              pointer = lblstack[i];
              achou = true;
              break;
            }
          }
          if (achou) {
            continue;
          } else {
            //printf("Label not found: %d\n", lbl);
            exit(-1);
          }
        }

      // comando: pop
      } else if (strcmp(argv[pointer], "POP") == 0) {
        restop--;

      // comando: prt
      } else if (strcmp(argv[pointer], "PRT") == 0) {
        int aux = restack[restop];
        char caux;
        while (aux) {
          caux = '0' + (aux % 10);
          write(pipe_fd[1], &caux, 1);
          aux /= 10;
        }
        caux = ' ';
        write(pipe_fd[1], &caux, 1);
        //printf("%d\n", restack[restop]);

      // operações aritméticas
      } else if (strcmp(argv[pointer], "+") == 0) {
        restack[restop - 1] = restack[restop - 1] + restack[restop];
        restop--;
      } else if (strcmp(argv[pointer], "-") == 0) {
        restack[restop - 1] = restack[restop - 1] - restack[restop];
        restop--;
      } else if (strcmp(argv[pointer], ".") == 0) {
        restack[restop - 1] = restack[restop - 1] * restack[restop];
        restop--;
      } else if (strcmp(argv[pointer], "/") == 0) {
        restack[restop - 1] = restack[restop - 1] / restack[restop];
        restop--;
      } else if (strcmp(argv[pointer], "%") == 0) {
        restack[restop - 1] = restack[restop - 1] % restack[restop];
        restop--;

      // atribuição de variáveis
      } else if (strcmp(argv[pointer], "LET") == 0) {
        // procura se o rótulo já existe
        bool achou = false;
        for (int i = 0; i < varcount; i++) {
          if (strcmp(varnstack[i], argv[pointer + 1]) == 0) {
            achou = true;
            varstack[i] = restack[restop];
            break;
          }
        }
        if (! achou) {
          varnstack = (char **)realloc(varnstack, (varcount + 1) * sizeof(char *));
          varnstack[varcount] = (char *)malloc( (1 + strlen(argv[pointer + 1])) * sizeof(char) );
          varnstack[varcount] = strcpy(varnstack[varcount], argv[pointer + 1]);

          varstack = (int *)realloc(varstack, (varcount + 1) * sizeof(int));
          varstack[varcount] = restack[restop];
          varcount++;
        }
        pointer++; // para pular o nome da variável

      // atribuição de arrays
      } else if (strcmp(argv[pointer], "ALET") == 0) {
        // procura se o rótulo já existe
        bool achou = false;
        for (int i = 0; i < arrcount; i++) {
          if (strcmp(arrnstack[i], argv[pointer + 1]) == 0) {
            achou = true;

            // verifica se a posição já está alocada
            if (atoi(argv[pointer + 2]) < arrcounter[i]) {
              // é só fazer a atribuição
              *(arrstack[i] + atoi(argv[pointer + 2])) = restack[restop];
            } else {
              // precisa abrir espaço no array para fazer a atribuição
              arrstack[i] = (int *)realloc(arrstack[i], (atoi(argv[pointer + 2]) + 1) * sizeof(int));
              *(arrstack[i] + atoi(argv[pointer + 2])) = restack[restop];
              arrcounter[i] = atoi(argv[pointer + 2]) + 1;
            }
            break;
          }
        }
        if (!achou) {
          arrnstack = (char **)realloc(arrnstack, (arrcount + 1) * sizeof(char *));
          arrnstack[arrcount] = (char *)malloc( (1 + strlen(argv[pointer + 1])) * sizeof(char) );
          arrnstack[arrcount] = strcpy(arrnstack[arrcount], argv[pointer + 1]);

          arrcounter = (int *)realloc(arrcounter, (arrcount + 1) * sizeof(int));
          arrcounter[arrcount] = atoi(argv[pointer + 2]) + 1;

          arrstack = (int **)realloc(arrstack, (arrcount + 1) * sizeof(int *));
          arrstack[arrcount] = (int *)realloc(arrstack[arrcount], (atoi(argv[pointer + 2]) + 1) * sizeof(int));
          *(arrstack[arrcount] + atoi(argv[pointer + 2])) = restack[restop];
          arrcount++;
        }
        pointer++; // para pular o nome da variável
        pointer++; // para pular o índice da variável

      // só pode ser o nome de uma variável, ou de um array
      } else {
        // procura o rótulo nas variáveis
        bool achou = false;
        for (int i = 0; i < varcount; i++) {
          if (strcmp(varnstack[i], argv[pointer]) == 0) {
            achou = true;
            // põe no topo da pilha de resultados
            if (restop >= rescount - 1) {
              rescount = restop + 2;
              restack = (int *)realloc(restack, rescount * sizeof(int));
            }
            restop++;
            restack[restop] = varstack[i];
            break;
          }
        }
        if (! achou) {
          // procura na pilha de arrays
          achou = false;
          for (int i = 0; i < arrcount; i++) {
            if (strcmp(arrnstack[i], argv[pointer]) == 0) {
              achou = true;

              // põe no topo da pilha de resultados (não checa se está dentro das boundaries: stops com segfault)
              // não precisa de aumentar a pilha de resultados porque vai substituir o topo da pilha pelo valor da variável
              //if (restop >= rescount - 1) {
              //  rescount = restop + 2;
              //  restack = (int *)realloc(restack, rescount * sizeof(int));
              //}
              // restop++;
              restack[restop] = *(arrstack[i] + restack[restop]);
              break;
            }
          }

          if (! achou) {
            // Erro, né
            //printf("Syntax error: %s\n", argv[pointer]);
            exit_status = -2;
            break;
          }
        }
      }
      pointer++;
    }
    //printf("FIM\n");

    free(restack);

    if (varstack) free(varstack);
    for (int i = 0; i < varcount; i++) {
      if (varnstack[i]) free(varnstack[i]);
    }
    free(varnstack);

    if (lblstack) free(lblstack);
    if (lblnstack) free(lblnstack);

    if (arrcounter) free(arrcounter);
    for (int i = 0; i < arrcount; i++) {
      if (arrnstack[i]) free(arrnstack[i]);
      if (arrstack[i]) free(arrstack[i]);
    }
    free(arrnstack);
    free(arrstack);

    close(pipe_fd[1]);
    exit(exit_status);

  } else if (pid > 0) {
    int status = 0;
    //printf("- right before wait()\n");

    int actualbuffsize = GAZ_BUFFSIZE;
    char *result;
    result = (char *)malloc(actualbuffsize * sizeof(char));
    int buffpointer = 0;

    // close write-end of pipe
    close(pipe_fd[1]);

    // reads results
    while (read(pipe_fd[0], &pipe_buffer, 1) > 0) {
      result[buffpointer] = pipe_buffer;
      buffpointer++;
      if (buffpointer >= actualbuffsize) {
        actualbuffsize += GAZ_BUFFSIZE;
        result = (char *)realloc(result, actualbuffsize * sizeof(char));
      }
    }
    result[buffpointer] = '\0';

    int wval = wait(&status);

    if (wval < 0) {
      //perror("Error in wait");
    } else {
      if (WIFEXITED(status)) {
        retval = true;
        //printf("-- child process %d exited normally with status %d. Result is '%s'\n", wval, WEXITSTATUS(status), result);
        if (return_status) *return_status = status;
        if (le_result) {
          if (strlen(result) >= GAZ_BUFFSIZE) {
            *le_result = (char *)realloc(*le_result, (strlen(result) + 1) * sizeof(char));
          }
          *le_result = strcpy(*le_result, result);
        }
      } else if (WIFSIGNALED(status)) {
        //printf("-- child process %d exited abnormally with signal %d\n", wval, WTERMSIG(status));
        if (return_status) *return_status = status;
      } else {
        //printf("--child process %d exited with a not treated condition\n", wval);
      }

      free(result);
    }
    //printf("- right after wait(), wval = %d, status = %d\n", wval, WEXITSTATUS(status));
  } else {
    //perror("fork() error!!!");
  }
  return retval;
}

